use std::sync::Arc;

use axum::{
    body::Body,
    handler::Handler,
    http::{Request, StatusCode},
    Extension, Router,
};
use config::Config;
use tower_http::{cors::CorsLayer, trace::TraceLayer};

mod api;
mod config;
mod models;
mod responses;
mod ruma_wrapper;
mod types;

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    if std::env::var("RUST_LOG").is_err() {
        std::env::set_var("RUST_LOG", "debug");
    }

    tracing_subscriber::fmt::init();

    let config = Arc::new(Config::default());

    let pool = sqlx::SqlitePool::connect(config.db_path()).await?;

    let cors = CorsLayer::new()
        .allow_origin(tower_http::cors::Any)
        .allow_methods(tower_http::cors::Any)
        .allow_headers(tower_http::cors::Any);

    let tracing_layer = TraceLayer::new_for_http();

    let client_server = Router::new()
        .merge(api::client_server::versions::routes())
        .merge(api::client_server::r0::auth::routes())
        .merge(api::client_server::r0::thirdparty::routes())
        .merge(api::client_server::r0::create_room::routes());

    let router = Router::new()
        .nest("/_matrix/client", client_server)
        .layer(cors)
        .layer(tracing_layer)
        .layer(Extension(pool))
        .layer(Extension(config))
        .fallback(fallback.into_service());

    let _ = axum::Server::bind(&"127.0.0.1:3000".parse().unwrap())
        .serve(router.into_make_service())
        .await?;

    Ok(())
}

async fn fallback(request: Request<Body>) -> StatusCode {
    tracing::error!("{} {}", request.method(), request.uri());
    StatusCode::INTERNAL_SERVER_ERROR
}
