use crate::types::uuid::Uuid;
use argon2::{password_hash::SaltString, Argon2, PasswordHash, PasswordHasher, PasswordVerifier};
use rand::rngs::OsRng;
use ruma::OwnedUserId;
use sqlx::SqlitePool;

#[derive(Debug)]
pub struct User {
    pub uuid: Uuid,
    pub user_id: String,
    pub display_name: String,
    pub password_hash: String,
}

impl User {
    pub fn new(user_id: &OwnedUserId, display_name: &str, password: &str) -> anyhow::Result<Self> {
        let argon2 = Argon2::default();
        let salt = SaltString::generate(OsRng);
        let password_hash = argon2
            .hash_password(password.as_bytes(), &salt)?
            .to_string();

        Ok(Self {
            uuid: uuid::Uuid::new_v4().into(),
            user_id: user_id.to_string(),
            display_name: display_name.to_owned(),
            password_hash,
        })
    }

    pub async fn exists(conn: &SqlitePool, user_id: &OwnedUserId) -> anyhow::Result<bool> {
        let user_id = user_id.to_string();
        Ok(
            sqlx::query!("select user_id from users where user_id = ?", user_id)
                .fetch_optional(conn)
                .await?
                .is_some(),
        )
    }

    pub async fn create(&self, conn: &SqlitePool) -> anyhow::Result<Self> {
        Ok(sqlx::query_as!(
            Self,
            "insert into users(uuid, user_id, display_name, password_hash)
            values (?, ?, ?, ?)
            returning uuid as 'uuid: Uuid', user_id, display_name, password_hash",
            self.uuid,
            self.user_id,
            self.display_name,
            self.password_hash
        )
        .fetch_one(conn)
        .await?)
    }

    pub async fn update(&self, conn: &SqlitePool) -> anyhow::Result<Self> {
        Ok(sqlx::query_as!(
            Self,
            "update users set uuid = ?, user_id = ?, display_name = ?, password_hash = ?
            where uuid = ?
            returning uuid as 'uuid: Uuid', user_id, display_name, password_hash",
            self.uuid,
            self.user_id,
            self.display_name,
            self.password_hash,
            self.uuid
        )
        .fetch_one(conn)
        .await?)
    }

    pub async fn find_by_uuid(conn: &SqlitePool, uuid: &Uuid) -> anyhow::Result<Self> {
        Ok(sqlx::query_as!(
            Self,
            "select uuid as 'uuid: Uuid', user_id, display_name, password_hash
             from users where uuid = ?",
            uuid
        )
        .fetch_one(conn)
        .await?)
    }

    pub async fn find_by_user_id(conn: &SqlitePool, user_id: &str) -> anyhow::Result<Self> {
        Ok(sqlx::query_as!(
            Self,
            "select uuid as 'uuid: Uuid', user_id, display_name, password_hash
             from users where user_id = ?",
            user_id
        )
        .fetch_one(conn)
        .await?)
    }

    pub fn password_correct(&self, password: &str) -> anyhow::Result<bool> {
        let password_hash = PasswordHash::new(&self.password_hash)?;

        Ok(Argon2::default()
            .verify_password(password.as_bytes(), &password_hash)
            .is_ok())
    }
}
